import numpy as np
import sys


def nrCRCEncode(blk, poly, mask=0):
    if poly == "6":
        L = 6
        # fmt: off
        crc_coeffs = [6, 5, 0]
        # fmt: on
    elif poly == "11":
        L = 11
        # fmt: off
        crc_coeffs = [11, 10, 9, 5, 0]
        # fmt: on
    elif poly == "16":
        L = 16
        # fmt: off
        crc_coeffs = [16, 12, 5, 0]
        # fmt: on
    elif poly == "24A":
        L = 24
        # fmt: off
        crc_coeffs = [24, 23, 18, 17, 14, 11, 10, 7, 6, 5, 4, 3, 1, 0]
        # fmt: on
    elif poly == "24B":
        L = 24
        # fmt: off
        crc_coeffs = [24, 23, 6, 5, 1, 0]
        # fmt: on
    elif poly == "24C":
        L = 24
        # fmt: off
        crc_coeffs = [24, 23, 21, 20, 17, 15, 13, 12, 8, 4, 2, 1, 0]
        # fmt: on
    else:
        print("Error: invalid CRC polynomial specified!")
        sys.exit()
    blksrc = np.empty(len(blk) + L, "int")
    blksrc[: len(blk)] = blk

    # build crc_poly
    crc_poly = 0
    for crc_coeff in crc_coeffs:
        crc_poly += 2**crc_coeff

    # calculate crc
    crc = 0
    for bit in blk:
        if ((crc >> (L - 1)) & 1) != bit:
            crc = (crc << 1) ^ crc_poly
        else:
            crc = crc << 1
    crc ^= mask

    # append crc LSB first
    for i in range(L):
        blksrc[len(blk) + i] = (crc >> (L - i - 1)) & 1
    return blksrc


def test_nrCRCEncode():
    test = np.zeros(32)
    test[0] = 1
    result = nrCRCEncode(test, "24C")
    # fmt: off
    assert np.all(result == [1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 0, 0, 0, 1, 1, 0, 0, 0, 1, 1, 0, 0, 0, 1, 1, 0, 0, 0, 0, 1, 0])
    # fmt: on



import numpy as np
import scipy
from py3gpp import nrSetResources


def nrChannelEstimate(rxGrid=None, refInd=None, refSym=None, refGrid=None, carrier=None):
    cp = "normal" if carrier is None else "extended"
    if refGrid is None:
        if (refInd is None) or (refSym is None):
            print("Error: refInd and refSym need to be given when refGrid is None!")
            return
        refGrid = np.zeros(rxGrid.shape, "complex")
        nrSetResources(refInd, refGrid, refSym)
    else:
        pilot_idx = np.where(refGrid.ravel(order="F") != 0)
        refInd = pilot_idx[0]
        refSym = refGrid.ravel(order="F")[refInd]

    rxSym = rxGrid.ravel(order="F")[refInd]
    L = 14 if cp == "normal" else 12
    K = rxGrid.shape[0]  # number of subcarriers
    N = rxGrid.shape[1]  # number of symbols
    H = np.zeros(refGrid.shape, "complex")
    R, C = np.mgrid[:K, :N]
    normGrid = rxGrid / refGrid
    idx_nonzero = np.where(np.invert(np.isinf(normGrid.ravel(order="F"))))[0]
    normXYZ = np.column_stack(
        (C.ravel(order="F")[idx_nonzero], R.ravel(order="F")[idx_nonzero], normGrid.ravel(order="F")[idx_nonzero])
    )

    if True:
        # this one works best!
        for col in range(N):
            idx = np.where(np.logical_and(refInd >= col * K, refInd < (col + 1) * K))[0]
            if len(idx) > 0:
                sym_idx = refInd[idx] - col * K
                H[:, col] = np.interp(np.arange(K), sym_idx, rxSym[idx] / refSym[idx])

        # estimate noise variance
        mean = np.mean((np.abs(H[:, 1]), np.abs(H[:, 3])), axis=0)
        nVar = np.mean(np.abs(np.abs(H[:, 1]) - mean))

    elif False:
        H = scipy.interpolate.griddata(
            (normXYZ[:, 0].real, normXYZ[:, 1].real), normXYZ[:, 2], (C.ravel(order="F"), R.ravel(order="F"))
        ).reshape(refGrid.shape, order="F")
    elif False:
        # use RBFInterpolator because interp2d and griddata cannot extrapolate
        interpolator = scipy.interpolate.RBFInterpolator(
            np.array((normXYZ[:, 0].real, normXYZ[:, 1].real)).T, normXYZ[:, 2]
        )
        H = interpolator(np.vstack((C.ravel(order="F"), R.ravel(order="F"))).T).reshape(refGrid.shape, order="F")
    return H, nVar



from .helper import _calc_gold

# v: scrambling sequence phase, 0..7
# n: number of elements in output sequence
def nrPBCHPRBS(ncellid, v, n):
    seq = _calc_gold(ncellid, n + v * n)
    return seq[v * n :][:n]



import numpy as np

def nrEqualizeMMSE(rxSym, hest, nVar):
    csi = (np.abs(hest * np.conj(hest)) + nVar)
    # rxEq = rxSym / hest # zero forcing
    rxEq = rxSym * np.conj(hest) / csi
    return rxEq, csi


import numpy as np
import scipy
from py3gpp import nrOFDMModulate


def nrTimingEstimate(
    carrier=None,
    waveform=None,
    refGrid=None,
    nrb=None,
    scs=None,
    initialNSlot=None,
    refInd=None,
    refSym=None,
    CyclicPrefix="normal",
    Nfft=None,
    SampleRate=None,
):
    refWaveform, _ = nrOFDMModulate(
        grid=refGrid, scs=scs, initialNSlot=initialNSlot, SampleRate=SampleRate, Nfft=Nfft, carrier=carrier
    )
    xcorr = scipy.signal.correlate(waveform, refWaveform, "valid")
    index = np.argmax(np.abs(xcorr))
    return index



import numpy as np

# TS 38.211 Section 5.2.1
def _calc_gold(c_init, length):
    Nc = 1600
    n = 2*length + 1 + 1600
    x1 = np.zeros(n)
    x1[0] = 1
    x2 = np.zeros(n)
    for i in range(30):
        x2[i] = (c_init>>i) & 1

    for i  in np.arange(n-31):
        x1[i + 31] = (x1[i+3] + x1[i])%2
        x2[i + 31] = (x2[i+3] + x2[i+2] + x2[i+1] + x2[i])%2

    c = np.empty(2*length+1, 'bool')
    for n in range(len(c)):
        c[n] = (x1[n+Nc] + x2[n+Nc])%2
    return c


import numpy as np


def nrOFDMInfo(carrier=None, nrb=None, scs=None):
    if nrb is None or scs is None:
        if carrier is None:
            print("Error: carrier is needed when no nrb or scs is given!")
            return
        nrb = carrier.NSizeGrid
        scs = carrier.SubcarrierSpacing

    for i in range(20):
        if 2 ** (i) * 0.85 >= nrb * 12:
            Nfft = 2**i
            break
    info = dict()
    info["Nfft"] = Nfft
    return info



import numpy as np

def nrPSS(ncellid):
    n = 127
    x = np.zeros(n, 'int32')    # m-sequence
    x[0:7] = np.array([0, 1, 1, 0, 1, 1, 1])
    for i in np.arange(0, 127 - 7):
        x[i+7] = (x[i + 4] + x[i]) % 2
    d_PSS = np.zeros(n, 'int32')
    for n in np.arange(0, 127):
        m = (n + 43*ncellid) % 127  # offset based on ncellid
        d_PSS[n] = 1 - 2*x[m]  # BPSK modulation
    return d_PSS


import numpy as np
from .helper import _calc_gold


def nrPBCHDMRS(ncellid, ibar_SSB):
    # ibar_SSB is a 3 bit value
    c_init = 2**11 * (ibar_SSB + 1) * (ncellid // 4 + 1) + 2**6 * (ibar_SSB + 1) + (ncellid % 4)
    c = _calc_gold(c_init, 144)

    r = np.empty(144, "complex")
    for m in range(len(r)):
        r[m] = 1 / np.sqrt(2) * (1 - 2 * c[2 * m]) + 1j * 1 / np.sqrt(2) * (1 - 2 * c[2 * m + 1])
    return r



import numpy as np


def nrPBCHDMRSIndices(ncellid, style="matlab"):
    indices_sym1_3 = np.arange(ncellid % 4, 240, 4)
    indices_sym2 = np.concatenate((np.arange(ncellid % 4, 48, 4), np.arange(192 + ncellid % 4, 240, 4)))
    if style == "python":
        indices = np.concatenate((indices_sym1_3, indices_sym2, indices_sym1_3))
        symbol_idx = np.concatenate(
            (1 * np.ones(len(indices_sym1_3)), 2 * np.ones(len(indices_sym2)), 3 * np.ones(len(indices_sym1_3)))
        ).astype("int")
        return (tuple(indices), tuple(symbol_idx))
    elif style == "matlab":
        indices = np.concatenate((indices_sym1_3 + 240, indices_sym2 + 2 * 240, indices_sym1_3 + 3 * 240))
        return indices
    else:
        print("Error: Unknown style!")
        return



from .nrPSS import nrPSS
from .nrSSS import nrSSS
from .nrPBCHDMRS import nrPBCHDMRS
from .nrCarrierConfig import nrCarrierConfig
from .nrResourceGrid import nrResourceGrid
from .nrOFDMModulate import nrOFDMModulate
from .nrOFDMDemodulate import nrOFDMDemodulate
from .nrSSSIndices import nrSSSIndices
from .nrExtractResources import nrExtractResources
from .nrPSSIndices import nrPSSIndices
from .nrTimingEstimate import nrTimingEstimate
from .nrPBCHDMRSIndices import nrPBCHDMRSIndices
from .nrSetResources import nrSetResources
from .nrChannelEstimate import nrChannelEstimate
from .nrOFDMInfo import nrOFDMInfo
from .nrSymbolDemodulate import nrSymbolDemodulate
from .nrPBCHIndices import nrPBCHIndices
from .nrPBCHPRBS import nrPBCHPRBS
from .nrEqualizeMMSE import nrEqualizeMMSE
from .nrRateRecoverPolar import nrRateRecoverPolar
from .nrPolarDecode import nrPolarDecode
from .nrCRCEncode import nrCRCEncode
from .nrCRCDecode import nrCRCDecode
from .nrBCHDecode import nrBCHDecode


import numpy as np
from py3gpp.nrCRCEncode import nrCRCEncode

def _binToInt(bin):
    integer = 0
    for bit in np.flip(bin):
        integer += bit
        integer <<= 1
    return integer

def nrCRCDecode(blkcrc, poly, mask=0):
    encoded = nrCRCEncode(blkcrc, poly, mask)
    L = len(encoded) - len(blkcrc)
    crc = encoded[-L:]
    return encoded[:len(blkcrc) - L], _binToInt(crc)

def test_nrCRCDecode():
    testdata = [0, 1, 0, 0, 1, 1, 1, 0, 0, 1, 1, 0, 1, 0, 0, 1, 1, 0, 0, 1, 1, 0, 1, 1, 1, 1, 1, 1, 0, 1, 1, 0, 0, 1, 1, 1, 0, 1, 0, 0, 1, 1, 0, 1, 1, 1, 1, 0, 1, 1, 1, 1, 1, 1, 1, 0]
    data, crc_result = nrCRCDecode(testdata, '24C')
    assert crc_result == 0

    testdata = [1, 1, 0, 0, 1, 1, 1, 0, 0, 1, 1, 0, 1, 0, 0, 1, 1, 0, 0, 1, 1, 0, 1, 1, 1, 1, 1, 1, 0, 1, 1, 0, 0, 1, 1, 1, 0, 1, 0, 0, 1, 1, 0, 1, 1, 1, 1, 0, 1, 1, 1, 1, 1, 1, 1, 0]
    data, crc_result = nrCRCDecode(testdata, '24C')
    assert crc_result != 0


import numpy as np
from py3gpp.nrOFDMInfo import nrOFDMInfo

# TODO: implement CyclicPrefixFraction
def nrOFDMDemodulate(
    carrier=None,
    waveform=None,
    nrb=None,
    scs=None,
    initialNSlot=None,
    CyclicPrefix="normal",
    Nfft=None,
    SampleRate=None,
    CarrierFrequency=0,
    CyclicPrefixFraction=0.5,
):
    if waveform is None:
        print("Error: no waveform given!")
        return
    if carrier is None:
        if nrb == None:
            print("Error: nrb is needed without carrierConfig!")
            return
        if scs == None:
            print("Error: scs is needed without carrierConfig!")
            return
        if initialNSlot == None:
            print("Error: initialNSlot is needed without carrierConfig!")
            return
    else:
        nrb = carrier.NSizeGrid
        scs = carrier.SubcarrierSpacing
        initialNSlot = 0
    if Nfft == None:
        if SampleRate == None:
            Nfft = nrOFDMInfo(nrb=nrb, scs=scs)["Nfft"]
            SampleRate = int(Nfft * scs * 1000)
        else:
            Nfft = int(SampleRate // scs // 1000)
    mu = (scs // 15) - 1
    if CyclicPrefix == "normal":
        N_cp1 = int(((144) * 2 ** (-mu) + 16) * (SampleRate / 30720000))
        N_cp2 = int((144 * 2 ** (-mu)) * (SampleRate / 30720000))
    else:
        N_cp1 = int((512 * 2 ** (-mu)) * (SampleRate / 30720000))
        N_cp2 = N_cp1

    idx = 0
    slot = 0
    grid = np.zeros((nrb * 12, 0), "complex")
    while idx + Nfft < waveform.shape[0]:
        if slot == 0 or slot == 7 * 2 ** (mu):
            cp = N_cp1
        else:
            cp = N_cp2
        idx += cp
        slot = (slot + 1) % (7 * 2 ** (mu))
        symbol_t = waveform[idx:][:Nfft]
        symbol_f = np.fft.fftshift(np.fft.fft(symbol_t))[Nfft // 2 - nrb * 12 // 2 : Nfft // 2 + nrb * 12 // 2]
        grid = np.concatenate((grid, np.expand_dims(symbol_f, 1)), axis=1)
        idx += Nfft
    return grid



import numpy as np
from py3gpp import nrCarrierConfig
from py3gpp.nrOFDMInfo import nrOFDMInfo

# TODO: implement windowing
def nrOFDMModulate(
    carrier=None, grid=None, scs=None, initialNSlot=0, CyclicPrefix="normal", Nfft=None, SampleRate=None, Windowing=None
):
    info = {}

    if carrier is None:
        if grid is None:
            print("Error: grid is needed when no carrierConfig is specified!")
            return
        NSizeGrid = grid.shape[0] // 12
        carrier = nrCarrierConfig(NSizeGrid=NSizeGrid)
    else:
        NSizeGrid = carrier.NSizeGrid

    if scs == None:
        scs = carrier.SubcarrierSpacing

    if Nfft == None:
        if SampleRate == None:
            Nfft = nrOFDMInfo(nrb=NSizeGrid, scs=scs)["Nfft"]
            SampleRate = int(Nfft * scs * 1000)
        else:
            Nfft = int(SampleRate // scs // 1000)

    mu = (scs // 15) - 1

    info["Nfft"] = Nfft
    info["SampleRate"] = SampleRate
    info["CyclicPrefixLengths"] = np.empty(0, "int")
    info["SymbolLengths"] = np.empty(0, "int")
    if len(grid.shape) == 1:
        grid = grid[..., np.newaxis]
    nSlots = grid.shape[1]
    waveform = np.empty(0, "complex")
    if CyclicPrefix == "normal":
        N_cp1 = int(((144) * 2 ** (-mu) + 16) * (SampleRate / 30720000))
        N_cp2 = int((144 * 2 ** (-mu)) * (SampleRate / 30720000))
    else:
        N_cp1 = int((512 * 2 ** (-mu)) * (SampleRate / 30720000))
        N_cp2 = N_cp1

    for slot_num in range(nSlots):
        if slot_num == 0 or slot_num == 7 * 2 ** (mu):
            N_cp = N_cp1
        else:
            N_cp = N_cp2
        symbol_len = Nfft + N_cp
        nFill = (Nfft - grid.shape[0]) // 2
        if nFill < 0:
            full_slot_grid = grid[np.abs(nFill) :, slot_num][:nFill]
        else:
            full_slot_grid = np.concatenate([np.zeros(nFill), grid[:, slot_num], np.zeros(nFill)])
        symbol_waveform = np.fft.ifft(np.fft.fftshift(full_slot_grid))
        symbol_waveform_cp = np.append(symbol_waveform[-N_cp:], symbol_waveform)
        waveform = np.append(waveform, symbol_waveform_cp)
        info["CyclicPrefixLengths"] = np.append(info["CyclicPrefixLengths"], N_cp)
        info["SymbolLengths"] = np.append(info["SymbolLengths"], symbol_len)
    return [waveform, info]



class nrCarrierConfig:
    def __init__(
        self, NCellId=1, SubcarrierSpacing=15, CyclicPrefix="normal", NSizeGrid=52, NStartGrid=0, NSlot=0, NFrame=0
    ):
        self.NCellId = NCellId
        self.SubcarrierSpacing = SubcarrierSpacing
        self.NSizeGrid = NSizeGrid
        self.NStartGrid = NStartGrid
        self.NSlot = NSlot
        self.NFrame = NFrame



import numpy as np


def nrResourceGrid(carrier, p=1):
    SymbolsPerSlot = 14
    if p == 1:
        grid = np.zeros((12 * carrier.NSizeGrid, SymbolsPerSlot), "complex")
    else:
        grid = np.zeros((12 * carrier.NSizeGrid, SymbolsPerSlot, p), "complex")
    return grid



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['py3gpp']

package_data = \
{'': ['*']}

install_requires = \
['numpy>=1.21.0,<2.0.0']

setup_kwargs = {
    'name': 'py3gpp',
    'version': '0.1.1',
    'description': 'functions for 5G NR signal processing',
    'long_description': "[![CI](https://github.com/catkira/py3gpp/actions/workflows/lint_and_test.yml/badge.svg)](https://github.com/catkira/py3gpp/actions/workflows/lint_and_test.yml)\n[![Pylint](https://catkira.github.io/py3gpp/pylint.svg)](https://github.com/catkira/py3gpp/actions/workflows/python-package.yml)\n[![PyPI version](https://badge.fury.io/py/py3gpp.svg)](https://badge.fury.io/py/py3gpp)\n[![Downloads](https://pepy.tech/badge/py3gpp)](https://pepy.tech/project/py3gpp)\n\n# Summary\nThis python package aims to replace the Matlab 5G Toolbox in Python. The call syntax of functions is the same as in matlab where possible. There are some differences, because matlab allows to continuously index a multidimensional array in one axis. In python this is not possible, therefore the result of functions like nrPBCHIndices() is also multidimensional here to make it compatible with Python.\n\n# Installation\n'python3 -m pip install py3gpp'\nor\nclone this repo and then do 'python3 -m pip install -e .'\n\n# Getting started\nrun 'examples/test_py3gpp.ipynb'\n\nThe example data is ideal data generated with Matlab, but the code has been tested with real data that has CFO, SFO and noise.\n\n# Documentation\nSee Matlab documentation of equivalent function\n",
    'author': 'Benjamin Menküc',
    'author_email': 'benjamin@menkuec.de',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'https://github.com/catkira/py3gpp',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
}


setup(**setup_kwargs)



import numpy as np

# 16QAM constellation:
#                Q
#  1011    1001  |   0001  0011
#  1010    1000  |   0000  0010
# ---------------------------------> I
#  1110    1100  |   0100  0110
#  1111    1101  |   0101  0111
def _16QAM_table():
    table = np.empty(16, "complex")
    QAM16_LEVEL_1 = 1 / np.sqrt(10)
    QAM16_LEVEL_2 = 3 / np.sqrt(10)
    table[0] = QAM16_LEVEL_1 + 1j * QAM16_LEVEL_1
    table[1] = QAM16_LEVEL_1 + 1j * QAM16_LEVEL_2
    table[2] = QAM16_LEVEL_2 + 1j * QAM16_LEVEL_1
    table[3] = QAM16_LEVEL_2 + 1j * QAM16_LEVEL_2
    table[4] = QAM16_LEVEL_1 - 1j * QAM16_LEVEL_1
    table[5] = QAM16_LEVEL_1 - 1j * QAM16_LEVEL_2
    table[6] = QAM16_LEVEL_2 - 1j * QAM16_LEVEL_1
    table[7] = QAM16_LEVEL_2 - 1j * QAM16_LEVEL_2
    table[8] = -QAM16_LEVEL_1 + 1j * QAM16_LEVEL_1
    table[9] = -QAM16_LEVEL_1 + 1j * QAM16_LEVEL_2
    table[10] = -QAM16_LEVEL_2 + 1j * QAM16_LEVEL_1
    table[11] = -QAM16_LEVEL_2 + 1j * QAM16_LEVEL_2
    table[12] = -QAM16_LEVEL_1 - 1j * QAM16_LEVEL_1
    table[13] = -QAM16_LEVEL_1 - 1j * QAM16_LEVEL_2
    table[14] = -QAM16_LEVEL_2 - 1j * QAM16_LEVEL_1
    table[15] = -QAM16_LEVEL_2 - 1j * QAM16_LEVEL_2
    return table


def nrSymbolDemodulate(input, mod, nVar=1e-10, DecisionType="soft"):
    output = np.empty(0, "float")

    for symbol in input:
        if mod == "BPSK":
            output = np.append(output, np.real(symbol) + np.imag(symbol))
        elif mod == "QPSK":
            output = np.append(output, np.real(symbol))
            output = np.append(output, np.imag(symbol))
        elif mod == "QAM16":
            output = np.append(output, np.real(symbol))
            output = np.append(output, np.imag(symbol))
            output = np.append(output, -(np.abs(np.real(symbol)) - 2 / np.sqrt(10)))
            output = np.append(output, -(np.abs(np.imag(symbol)) - 2 / np.sqrt(10)))

    if DecisionType == "soft":
        output /= nVar / np.exp(1)
        if mod == "QAM16":
            output /= 2
    else:
        output = output < 0
    return output



import numpy as np

# this function does not exist in Matlab, but its useful in python because
# numpy does not support integer indexing across dimensions
def nrSetResources(ind, grid, vals):
    nSymbols = grid.shape[0]
    for i in range(grid.shape[1]):
        idx = np.where(np.logical_and(ind >= i * nSymbols, ind < (i + 1) * nSymbols))[0]
        if len(idx) != 0:
            grid[ind[idx] - i * nSymbols, i] = vals[idx]
    return vals



import numpy as np
from py3gpp.nrPBCHDMRSIndices import nrPBCHDMRSIndices


def nrPBCHIndices(ibar):
    indices = np.hstack((np.arange(240, 480), np.arange(480, 528), np.arange(672, 720), np.arange(720, 960)))
    return np.setdiff1d(indices, nrPBCHDMRSIndices(ibar))
